from typing import Optional
import argparse
import sys
import os
from pathlib import Path

VERSION = "development"

def parse_args():
    parser = argparse.ArgumentParser(description="Generates Python bindings for a Pkl module")
    
    # Positional argument for module
    parser.add_argument('module', help="The module for which to generate bindings", nargs="?")

    # Optional arguments
    parser.add_argument('--generator-settings', help="The path to a generator settings file", default="")
    parser.add_argument('--output-path', help="The output directory to write generated sources into", default="")
    parser.add_argument('--base-path', help="The base path used to determine relative output", default="")
    parser.add_argument('--mapping', action='append', help="The mapping of a Pkl module name to a Python package name", default=[])
    parser.add_argument('--suppress-format-warning', action='store_true', help="Suppress warnings around formatting issues")
    parser.add_argument('--allowed-modules', action='append', help="URI patterns that determine which modules can be loaded and evaluated", default=[])
    parser.add_argument('--allowed-resources', action='append', help="URI patterns that determine which resources can be loaded and evaluated", default=[])
    parser.add_argument('--project-dir', help="The project directory to load dependency and evaluator settings from", default="")
    parser.add_argument('--dry-run', action='store_true', help="Print out the names of the files that will be generated, but don't write any files")
    # parser.add_argument('--version', action='store_true', help="Print the version and exit")

    return parser.parse_args()

def find_project_dir(start_path: str) -> Optional[str]:
    current_path = Path(start_path)
    while current_path != current_path.parent:
        if (current_path / "PklProject").exists():
            return str(current_path)
        current_path = current_path.parent
    return None


// Loads the settings for controlling codegen.
// Uses a Pkl evaluator that is separate from what's used for actually running codegen.
func loadGeneratorSettings(generatorSettingsPath string, projectDirFlag string) (*generatorsettings.GeneratorSettings, error) {
	projectDir := findProjectDir(projectDirFlag)
	var evaluator pkl.Evaluator
	var err error
	if projectDir != "" {
		evaluator, err = pkl.NewProjectEvaluator(context.Background(), projectDir, pkl.PreconfiguredOptions)
	} else {
		evaluator, err = pkl.NewEvaluator(context.Background(), pkl.PreconfiguredOptions)
	}
	if err != nil {
		panic(err)
	}
	var source *pkl.ModuleSource
	if generatorSettingsPath != "" {
		source = pkl.FileSource(generatorSettingsPath)
	} else if fileExists("generator-settings.pkl") {
		source = pkl.FileSource("generator-settings.pkl")
	} else {
		source = generatorSettingsSource()
	}
	return generatorsettings.Load(context.Background(), evaluator, source)
}

def load_generator_settings(generator_settings_path, project_dir_flag):
    project_dir = find_project_dir(project_dir_flag)
    evaluator = None  # Placeholder for evaluator initialization
    # Initialize your evaluator here based on project_dir, if applicable

    settings_path = generator_settings_path if generator_settings_path else "generator-settings.pkl"
    if not os.path.exists(settings_path):
        # Fallback to a default source or handle error
        print("Settings file not found.")
        return None

    with open(settings_path, 'r') as settings_file:
        settings = json.load(settings_file)
        # Assuming the settings file is in JSON format
        # For YAML, use `settings = yaml.safe_load(settings_file)`

    # Placeholder: Load and return the settings using the evaluator
    # This part of the logic will depend on how you intend to use the settings and the evaluator
    return settings


def main():
    args = parse_args()

    if not args.module:
        print("Error: Module name is required.")
        sys.exit(1)

    project_dir = find_project_dir(os.getcwd())
    evaluator = PklEvaluator(project_dir)
    settings = {}
    if args.generator_settings:
        settings = load_generator_settings(args.generator_settings, project_dir)

    # Placeholder for the main logic to generate Python bindings.
    # You would need to implement this based on how the Python bindings are generated from Pkl files.
    print(f"Generating bindings for module: {args.module} with settings: {settings} and output path: {args.output_path}")

if __name__ == "__main__":
    main()



# Code generated from Pkl module `lib3`. DO NOT EDIT.
from enum import Enum
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll

_NO_DEFAULT = object()

@dataclass
class GoGoGo:
    duck: Literal["quack"] = "quack"

    _registered_identifier: str = "lib3#GoGoGo"

@dataclass
class lib3:
    _registered_identifier: str = "lib3"

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `lib3.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())


import argparse
from pathlib import Path

import requests

VERSION = "0.25.2"

BASE_PATH = "https://github.com/apple/pkl/releases/download/"
filenames = [
    "pkl-macos-aarch64",
    "pkl-macos-amd64",
    "pkl-linux-aarch64",
    "pkl-linux-amd64",
    "pkl-alpine-linux-amd64",
]
JAVA_FILE = f"https://repo1.maven.org/maven2/org/pkl-lang/pkl-cli-java/{VERSION}/pkl-cli-java-{VERSION}.jar"


def main(save_path):
    urls = [BASE_PATH + VERSION + "/" + file for file in filenames]
    urls.append(JAVA_FILE)

    for url in urls:
        file = url.split("/")[-1]
        fp = save_path / file

        response = requests.get(url)
        with open(fp, "wb") as f:
            f.write(response.content)

        print("written to", fp)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "output_path", help="Directory where binaries will be downloaded"
    )

    args = parser.parse_args()

    main(Path(args.output_path))



import os
from pathlib import Path
from tempfile import TemporaryDirectory
import sys
import pkll

# Assuming 'Evaluator' and 'GeneratorSettings' are classes you will define or adapt in Python
# from src.evaluator.evaluator import Evaluator
# from generated import GeneratorSettings
import logging

logger = logging.getLogger()
logging.basicConfig(level=logging.INFO, format="%(message)s")
chalk = logging  # Simplified representation; in Python, you might use colorama or termcolor for colored output

class GeneratorSettings:
    """
    GeneratorSettings holds configuration for generating TypeScript files from PKL modules.
    Attributes:
        output_directory (str|None): The output path to write generated files into.
        dry_run (bool|None): If true, evaluates the PKL modules but does not write any files.
        generator_script_path (str|None): The Generator.pkl script to use for code generation.
    """
    def __init__(self, output_directory=None, dry_run=None, generator_script_path=None):
        self.output_directory = output_directory
        self.dry_run = dry_run
        self.generator_script_path = generator_script_path


def to_absolute_path(path: str) -> Path:
    return Path(path).resolve()


def generate_python(
    pkl_module_paths: list, settings: GeneratorSettings, verbose: bool,
):
    logger.info(
        f"Generating TypeScript sources for modules {', '.join(pkl_module_paths)}"
    )

    pkl_module_paths = [to_absolute_path(path) for path in pkl_module_paths]

    if settings.generator_script_path:
        if ":" in settings.generator_script_path:
            settings.generator_script_path = Path(settings.generator_script_path)
        else:
            settings.generator_script_path = to_absolute_path(
                settings.generator_script_path
            )
        logger.warning(
            f"Using custom generator script: {settings.generator_script_path}"
        )
    else:
        # Adjust the path as necessary for your project structure
        settings.generator_script_path = (
            Path(__file__).parent / "../codegen/src/Generator.pkl"
        )

    with TemporaryDirectory() as tmp_dir:
        output_dir = to_absolute_path(
            settings.output_directory if settings.output_directory else ".out"
        )
        output_dir.mkdir(parents=True, exist_ok=True)

        for index, pkl_input_module in enumerate(pkl_module_paths):
            module_to_evaluate = f"""
amends "{settings.generator_script_path}"

import "{pkl_input_module}" as theModule

moduleToGenerate = theModule
            """

            if logger.getEffectiveLevel() <= logging.DEBUG:
                logger.info(
                    f"""
Evaluating temp Pkl module:
---
{module_to_evaluate}
                """
                )

            tmp_file_path = Path(tmp_dir) / f"pkl-gen-python-{index}.pkl"
            with open(tmp_file_path, "w", encoding="utf-8") as tmp_file:
                tmp_file.write(module_to_evaluate)

            files = pkll.load(tmp_file_path.as_uri(), debug=verbose)

            for filename, contents in files.items():
                path = output_dir / filename
                if not settings.dry_run:
                    with open(path, "w", encoding="utf-8") as file:
                        file.write(contents)
                logger.info(f"Generated: {path}")


# Mockup classes to complete the example
class Evaluator:
    async def evaluate_output_files(self, uri):
        return {"example.ts": "const example = true;"}


# Example usage
def main():
    settings = GeneratorSettings(output_directory="path/to/output", dry_run=False)
    generate_python(
        ["path/to/module1", "path/to/module2"], settings
    )


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse

from pkll.evaluator import EVALUATOR_DEFAULT, Evaluator
from pkll.msgapi.outgoing import (
    ClientModuleReader,
    ClientResourceReader,
    Project,
    RemoteDependency,
)
from pkll.server import PKLServer

# get version
with open(os.path.join(os.path.dirname(__file__), "VERSION"), "r") as _f:
    __version__ = _f.read().strip()


def _get_project(project_conf):
    name = project_conf.__class__.__name__

    if name == "Project":
        return Project(
            projectFileUri=project_conf.projectFileUri,
            packageUri=project_conf.package.uri
            if project_conf.package is not None
            else None,
            dependencies={
                k: _get_project(v) for k, v in project_conf.dependencies.items()
            },
        )
    elif name == "RemoteDependency":
        return RemoteDependency(packageUri=project_conf.packageUri)
    else:
        raise ValueError(f"Unknown dependency: '{name}'")


def _search_project_dir(module_path: Path, debug=False) -> Project:
    cur_path = module_path
    while not (cur_path / "PklProject").exists():
        cur_path = cur_path.parent
        if str(cur_path) == "/":
            break

    if str(cur_path) == "/":
        cur_path = module_path

    cur_path = cur_path / "PklProject"
    if cur_path.exists():
        config = load(cur_path.as_uri(), project=None, debug=debug)
        project = _get_project(config)
    else:
        project = Project(projectFileUri=cur_path.as_uri())

    # project = Project(packageUri=cur_path.as_uri(), projectFileUri=cur_path.as_uri())
    return project


def load(
    moduleUri: Union[str, Path],
    moduleText: Optional[str] = None,
    expr: Optional[str] = None,
    *,
    force_render=False,
    parser=None,
    allowedModules: Optional[List[str]] = [
        "pkl:",
        "file:",
        "modulepath:",
        "https:",
        "repl:",
        "package:",
        "projectpackage:",
    ],
    allowedResources: Optional[List[str]] = [
        "env:",
        "prop:",
        "package:",
        "projectpackage:",
    ],
    clientModuleReaders: Optional[List[ClientModuleReader]] = None,
    clientResourceReaders: Optional[List[ClientResourceReader]] = None,
    modulePaths: Optional[List[str]] = None,
    env: Optional[Dict[str, str]] = EVALUATOR_DEFAULT,
    properties: Optional[Dict[str, str]] = None,
    timeoutSeconds: Optional[int] = None,
    rootDir: Optional[str] = None,
    cacheDir: Optional[str] = EVALUATOR_DEFAULT,
    outputFormat: Optional[str] = None,
    project: Optional[Project] = EVALUATOR_DEFAULT,
    module_handler=None,
    resource_handler=None,
    debug=False,
    **kwargs,
):
    """
    Loads and evaluates a Pkl module or expression with specified parameters and customization options.

    Args:
        module_uri (str): The absolute URI of the module to be loaded.
        module_text (Optional[str], None): Optionally, the content of the module to be loaded.
            If None, the module is loaded from the specified URI.
        expr (Optional[str], None): Optionally, a Pkl expression to be evaluated
            within the loaded module. If None, the entire module is evaluated.
        force_render (bool, False): Force the rendering of the module,
            even if it might not be necessary under normal circumstances.
        parser: A specific parser to be used for parsing the module.
            If None, a default parser is used.
        allowedModules (Optional[List[str]]): List of URI schemes that are allowed
            for module imports, with default schemes provided.
        allowedResources (Optional[List[str]]): List of URI schemes that are allowed
            for resource reading, with default schemes provided.
        clientModuleReaders (Optional[List[ClientModuleReader]]): Custom module readers
            for handling specific URI schemes on the client side.
        clientResourceReaders (Optional[List[ClientResourceReader]]): Custom resource
            readers for handling specific URI schemes on the client side.
        modulePaths (Optional[List[str]]): Additional paths to search when resolving module URIs.
        env (Optional[Dict[str, str]], EVALUATOR_DEFAULT): Environment variables to set
            for the evaluation process.
        properties (Optional[Dict[str, str]], None): External properties to set
            for the evaluation process.
        timeoutSeconds (Optional[int], None): Maximum duration in seconds
            for the evaluation to complete.
        rootDir (Optional[str], None): Restrict file-based module and resource access
            to within this root directory.
        cacheDir (Optional[str], EVALUATOR_DEFAULT): Directory to use for caching packages.
        outputFormat (Optional[str], None): Desired output format for the evaluation result.
        project (Optional[Project], None): Project dependencies and settings for the evaluation.
        module_handler: Custom handler for module loading and evaluation processes.
        resource_handler: Custom handler for resource loading processes.
        debug (bool, False): Enable debugging mode for additional output and diagnostics.
        **kwargs: Additional keyword arguments for extensibility and future use.

    Returns:
        The result of the module or expression evaluation, depending on the inputs and configuration.

    This function provides a flexible interface for loading and evaluating Pkl modules
    with a variety of customization options, including custom module and resource readers,
    environmental configurations, and support for complex project dependencies.
    """
    parsed = urlparse(str(moduleUri))
    default_scheme = "file"

    parsed = parsed._replace(
        scheme=parsed.scheme or default_scheme,
        path="" if parsed.path == "" else str(Path(parsed.path).absolute()),
    )
    moduleUri = parsed.geturl()

    if project is EVALUATOR_DEFAULT:
        project = _search_project_dir(Path(parsed.path).parent, debug=debug)

    with Evaluator(
        allowedModules=allowedModules,
        allowedResources=allowedResources,
        clientModuleReaders=clientModuleReaders,
        clientResourceReaders=clientResourceReaders,
        modulePaths=modulePaths,
        env=env,
        properties=properties,
        timeoutSeconds=timeoutSeconds,
        rootDir=rootDir,
        cacheDir=cacheDir,
        outputFormat=outputFormat,
        project=project,
        debug=debug,
        **kwargs,
    ) as evaluator:
        return evaluator.request(
            moduleUri,
            moduleText,
            expr,
            force_render=force_render,
            parser=parser,
            module_handler=module_handler,
            resource_handler=resource_handler,
        )


__all__ = ["load", "Evaluator", "PKLServer"]



# Code generated from Pkl module `Foo`. DO NOT EDIT.
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll

_NO_DEFAULT = object()

@dataclass
class Animal(Being):
    name: str = _NO_DEFAULT

    _registered_identifier: str = "Foo#Animal"

    def __post_init__(self):
        no_default_list = [("name", self.name)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

@dataclass
class Being:
    exists: bool

    _registered_identifier: str = "Foo#Being"

@dataclass
class Bird(Animal):
    flies: bool = _NO_DEFAULT

    _registered_identifier: str = "Foo#Bird"

    def __post_init__(self):
        no_default_list = [("flies", self.flies)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

@dataclass
class Dog(Animal):
    barks: bool = _NO_DEFAULT

    _registered_identifier: str = "Foo#Dog"

    def __post_init__(self):
        no_default_list = [("barks", self.barks)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

@dataclass
class Foo:
    animals: List[Animal]

    _registered_identifier: str = "Foo"

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `Foo.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())


# Code generated from Pkl module `MyModule`. DO NOT EDIT.
from enum import Enum
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll

_NO_DEFAULT = object()

@dataclass
class MyModule:
    foo: str

    _registered_identifier: str = "MyModule"


# Code generated from Pkl module `union`. DO NOT EDIT.
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll

_NO_DEFAULT = object()

City = Literal["San Francisco", "London", "上海"]

County = Literal["San Francisco", "San Mateo", "Yolo"]

Noodles = Literal["拉面", "刀切面", "面线", "意大利面"]

AccountDisposition = Literal["", "icloud3", "prod", "shared"]

@dataclass
class union:
    # A city
    city: City

    # County
    county: County

    # Noodles
    noodle: Noodles

    # Account disposition
    disposition: AccountDisposition

    _registered_identifier: str = "union"

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `union.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())


# Code generated from Pkl module `Imports`. DO NOT EDIT.
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll
from . import Foo_pkl

_NO_DEFAULT = object()

@dataclass
class Imports:
    foo: Foo_pkl.Foo

    _registered_identifier: str = "Imports"

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `Imports.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())


# Code generated from Pkl module `ExtendModule`. DO NOT EDIT.
from enum import Enum
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll
from . import MyModule_pkl

_NO_DEFAULT = object()

@dataclass
class ExtendModule(MyModule_pkl.MyModule):
    bar: str = _NO_DEFAULT

    _registered_identifier: str = "ExtendModule"

    def __post_init__(self):
        no_default_list = [("bar", self.bar)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `ExtendModule.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())

breakpoint()



import os
import warnings
from collections import deque
from typing import Dict, List, Optional

import msgpack

from pkll.handler import ModuleHandler, ResourcesHandler
from pkll.msgapi.code import (
    CODE_EVALUATE_LOG,
    CODE_EVALUATE_READ,
    CODE_EVALUATE_READ_MODULE,
    CODE_EVALUATE_RESPONSE,
    CODE_LIST_MODULES_REQUEST,
    CODE_LIST_RESOURCES_REQUEST,
)
from pkll.msgapi.incoming import (
    CreateEvaluatorResponse,
    EvaluateResponse,
    EvaluatorListModulesRequest,
    EvaluatorListResourcesRequest,
    EvaluatorReadModuleRequest,
    EvaluatorReadResourceRequest,
    Log,
)
from pkll.msgapi.outgoing import (
    ClientModuleReader,
    ClientResourceReader,
    CloseEvaluator,
    CreateEvaluator,
    EvaluateRequest,
    EvaluatorListModulesResponse,
    EvaluatorListResourcesResponse,
    EvaluatorReadModuleResponse,
    EvaluatorReadResourceResponse,
    Project,
)
from pkll.parser import Parser
from pkll.server import PKLServer

EVALUATOR_DEFAULT = "DEFAULT"


class Evaluator:
    def __init__(
        self,
        allowedModules: Optional[List[str]] = [
            "pkl:",
            "file:",
            "modulepath:",
            "https:",
            "repl:",
            "package:",
            "projectpackage:",
        ],
        allowedResources: Optional[List[str]] = [
            "env:",
            "prop:",
            "package:",
            "projectpackage:",
        ],
        clientModuleReaders: Optional[List[ClientModuleReader]] = None,
        clientResourceReaders: Optional[List[ClientResourceReader]] = None,
        modulePaths: Optional[List[str]] = None,
        env: Optional[Dict[str, str]] = EVALUATOR_DEFAULT,
        properties: Optional[Dict[str, str]] = None,
        timeoutSeconds: Optional[int] = None,
        rootDir: Optional[str] = None,
        cacheDir: Optional[str] = EVALUATOR_DEFAULT,
        outputFormat: Optional[str] = None,
        project: Optional[Project] = None,
        debug=False,
    ):
        self._server = PKLServer()
        self._evaluator_id = None

        if env == EVALUATOR_DEFAULT:
            env = dict(os.environ)

        if cacheDir == EVALUATOR_DEFAULT:
            dirname = os.path.expanduser("~")
            defaultCacheDir = os.path.join(dirname, ".pkl", "cache")
            cacheDir = defaultCacheDir

        self._CreateEvaluatorOpts = CreateEvaluator(
            requestId=self._server.get_request_id(),
            allowedModules=allowedModules,
            allowedResources=allowedResources,
            clientModuleReaders=clientModuleReaders,
            clientResourceReaders=clientResourceReaders,
            modulePaths=modulePaths,
            env=env,
            properties=properties,
            timeoutSeconds=timeoutSeconds,
            rootDir=rootDir,
            cacheDir=cacheDir,
            outputFormat=outputFormat,
            project=project,
        )
        self._debug = debug

    def create(self):
        self._server.start_process(debug=self._debug)

        msg_obj = self._CreateEvaluatorOpts.to_msg_obj()
        responses = self._handle_send_and_receive(msg_obj)
        assert len(responses) == 1
        code, msg_body = responses[0]

        response = CreateEvaluatorResponse(**msg_body)
        self._evaluator_id = response.evaluatorId
        return response

    def close(self):
        if self._evaluator_id is None:
            raise ValueError("Evaluator not created")

        msg_obj = CloseEvaluator(self._evaluator_id).to_msg_obj()
        _ = self._server.send_and_receive(msg_obj)

    def terminate(self):
        self._server.terminate()

    def _emit_logs(self, responses):
        ex_log = []

        for res in responses:
            is_log = res[0] == CODE_EVALUATE_LOG
            if is_log:
                self._handle_log_message(res)
            else:
                ex_log.append(res)

        return ex_log

    def _handle_send_and_receive(self, msg_obj, max_retry=5) -> List:
        responses = self._server.send_and_receive(msg_obj)
        responses = self._emit_logs(responses)

        count = 0
        while len(responses) == 0 and count < max_retry:
            responses = self._server.receive_with_retry(max_retry=1)
            responses = self._emit_logs(responses)
            count += 1
        return responses

    def _handle_log_message(self, response: List):
        code, msg_body = response
        response = Log(**msg_body)
        level_map = {
            0: "TRACE",
            1: "WARN",
        }
        msg = f"pkl: {level_map[response.level]}: {response.message} ({response.frameUri})"
        if response.level == 0:
            print(msg)
        elif response.level == 1:
            warnings.warn(msg)

    def request(
        self,
        moduleUri: str,
        moduleText=None,
        expr=None,
        force_render=False,
        parser=None,
        module_handler: Optional[ModuleHandler] = None,
        resource_handler: Optional[ResourcesHandler] = None,
    ):
        parser = parser or Parser(force_render=force_render)

        msg_obj = EvaluateRequest(
            requestId=self._server.get_request_id(),
            evaluatorId=self._evaluator_id,
            moduleUri=moduleUri,
            moduleText=moduleText,
            expr=expr,
        ).to_msg_obj()

        all_responses = self._handle_send_and_receive(msg_obj)

        queue = deque()
        queue.extend(all_responses)
        results = []

        while queue:
            code, msg_body = queue.popleft()
            if code == CODE_EVALUATE_RESPONSE:
                response = EvaluateResponse(**msg_body)

                if response.error:
                    raise Exception(response.error)
                raw = response.result
                result = msgpack.unpackb(raw, strict_map_key=False)
                parsed = parser.parse(result)
                results.append(parsed)
            elif code == CODE_LIST_MODULES_REQUEST:
                request = EvaluatorListModulesRequest(**msg_body)
                if module_handler is None:
                    raise ValueError("'module_handler' not given")
                res = module_handler.list_response(request.uri)
                msg_obj = EvaluatorListModulesResponse(
                    requestId=request.requestId,
                    evaluatorId=request.evaluatorId,
                    pathElements=res.pathElements,
                    error=res.error,
                ).to_msg_obj()
                all_responses = self._handle_send_and_receive(msg_obj)
                queue.extend(all_responses)
            elif code == CODE_LIST_RESOURCES_REQUEST:
                request = EvaluatorListResourcesRequest(**msg_body)
                if resource_handler is None:
                    raise ValueError("'resource_handler' not given")
                res = resource_handler.list_response(request.uri)
                msg_obj = EvaluatorListResourcesResponse(
                    requestId=request.requestId,
                    evaluatorId=request.evaluatorId,
                    pathElements=res.pathElements,
                    error=res.error,
                ).to_msg_obj()
                all_responses = self._handle_send_and_receive(msg_obj)
                queue.extend(all_responses)
            elif code == CODE_EVALUATE_READ_MODULE:
                request = EvaluatorReadModuleRequest(**msg_body)
                if module_handler is None:
                    raise ValueError("'module_handler' not given")
                res = module_handler.read_response(request.uri)
                msg_obj = EvaluatorReadModuleResponse(
                    requestId=request.requestId,
                    evaluatorId=request.evaluatorId,
                    contents=res.contents,
                    error=res.error,
                ).to_msg_obj()
                all_responses = self._handle_send_and_receive(msg_obj)
                queue.extend(all_responses)
            elif code == CODE_EVALUATE_READ:
                request = EvaluatorReadResourceRequest(**msg_body)
                if resource_handler is None:
                    raise ValueError("'resource_handler' not given")
                res = resource_handler.read_response(request.uri)
                msg_obj = EvaluatorReadResourceResponse(
                    requestId=request.requestId,
                    evaluatorId=request.evaluatorId,
                    contents=res.contents,
                    error=res.error,
                ).to_msg_obj()
                all_responses = self._handle_send_and_receive(msg_obj)
                queue.extend(all_responses)
            else:
                raise NotImplementedError(f"code: {hex(code)}")

        assert len(results) == 1
        return results[0]

    def __enter__(self):
        self.create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self.terminate()



from collections import namedtuple
from dataclasses import dataclass, make_dataclass
from datetime import timedelta
from enum import Enum, auto
from typing import Any, Literal


class ResultType(Enum):
    INT = auto()
    FLOAT = auto()
    STRING = auto()
    LIST = auto()
    TUPLE = auto()
    CLASS = auto()
    DICTIONARY = auto()
    RANGE = auto()
    DATACLASS = auto()
    NAMEDTUPLE = auto()
    TIMEDELTA = auto()
    PANDAS_TIMEDELTA = auto()


CODE_TYPED_DYNAMIC = 0x1
CODE_MAP = 0x2
CODE_MAPPING = 0x3
CODE_LIST = 0x4
CODE_LISTING = 0x5
CODE_SET = 0x6
CODE_DURATION = 0x7
CODE_DATASIZE = 0x8
CODE_PAIR = 0x9
CODE_INTSEQ = 0xA
CODE_REGEX = 0xB
CODE_CLASS = 0xC
CODE_TYPEALIAS = 0xD
CODE_PROPERTY = 0x10
CODE_ENTRY = 0x11
CODE_ELEMENT = 0x12


@dataclass
class Pair:
    first: Any
    second: Any


@dataclass
class Duration:
    value: float
    unit: Literal["ns", "us", "ms", "s", "min", "h", "d"]
    _UNIT_MAP = {
        "ns": "nanoseconds",
        "us": "microseconds",
        "ms": "milliseconds",
        "s": "seconds",
        "min": "minutes",
        "h": "hours",
        "d": "days",
    }

    def to_timedelta(self):
        return timedelta(**{self._UNIT_MAP[self.unit]: self.value})

    def to_pandas_timedelta(self):
        import pandas as pd

        return pd.Timedelta(self.value, self._UNIT_MAP[self.unit])


@dataclass
class DataSize:
    value: float
    unit: Literal["b", "kb", "kib", "mb", "mib", "gb", "gib", "tb", "tib", "pb", "pib"]


@dataclass
class Regex:
    pattern: str


@dataclass
class IntSeq:
    start: int
    end: int
    step: int


class Parser:
    def __init__(
        self,
        force_render=False,
        *,
        typed_dynamic_result_type=ResultType.DATACLASS,
        pair_result_type=ResultType.DATACLASS,
        datasize_result_type=ResultType.DATACLASS,
        regex_result_type=ResultType.DATACLASS,
        intseq_result_type=ResultType.DATACLASS,
    ):
        self.type_handlers = {
            CODE_TYPED_DYNAMIC: self.parse_typed_dynamic,
            CODE_MAP: self.parse_map,
            CODE_MAPPING: self.parse_mapping,
            CODE_LIST: self.parse_list,
            CODE_LISTING: self.parse_listing,
            CODE_SET: self.parse_set,
            CODE_DURATION: self.parse_duration,
            CODE_DATASIZE: self.parse_datasize,
            CODE_PAIR: self.parse_pair,
            CODE_INTSEQ: self.parse_intseq,
            CODE_REGEX: self.parse_regex,
            CODE_CLASS: self.parse_class,
            CODE_TYPEALIAS: self.parse_typealias,
            CODE_PROPERTY: self.parse_property,
            CODE_ENTRY: self.parse_entry,
            CODE_ELEMENT: self.parse_element,
        }
        self._dataclass_cache = {}
        self._namedtuple_cache = {
            "__Pair__": namedtuple("Pair", ["first", "second"]),
            "__DataSize__": namedtuple("DataSize", ["value", "unit"]),
            "__Regex__": namedtuple("Regex", ["pattern"]),
        }
        self._force_render = force_render
        self._typed_dynamic_result_type = typed_dynamic_result_type
        self._pair_result_type = pair_result_type
        self._datasize_result_type = datasize_result_type
        self._regex_result_type = regex_result_type
        self._intseq_result_type = intseq_result_type

    def parse(self, obj):
        return self.handle_type(obj)

    def handle_type(self, obj):
        if isinstance(obj, list):
            type_code = obj[0]
            if type_code in self.type_handlers:
                return self.type_handlers[type_code](obj)
            else:
                # Handle unknown type or implement a default handler
                return {"type": "Unknown", "value": obj}
        elif isinstance(obj, dict):
            return {k: self.handle_type(v) for k, v in obj.items()}
        elif isinstance(obj, (set, tuple)):
            return type(obj)(self.handle_type(v) for v in obj)
        return obj

    def parse_typed_dynamic(self, obj):
        _, full_class_name, module_uri, members = obj

        member_types = set(m[0] for m in members)
        property_list = list(map(self.handle_type, members))

        if CODE_ELEMENT in member_types:  # has element
            if len(member_types) > 1 and not self._force_render:
                raise ValueError(
                    "Cannot render object with both elements and properties/entries.\n"
                    "\tUse 'force_render=True'"
                )
            # element types
            members = property_list
            return members

        # only properties and entries
        members = {k: v for m in property_list for k, v in m.items()}
        result_type = self._typed_dynamic_result_type
        class_name = full_class_name.split("#")[-1].split(".")[-1]
        if result_type == ResultType.DATACLASS:
            res = make_dataclass(class_name, members.keys())(*members.values())
            return res
        elif result_type == ResultType.NAMEDTUPLE:
            res = namedtuple(class_name, members.keys())(*members.values())
            return res
        elif result_type == ResultType.DICTIONARY:
            return {full_class_name: members}
        else:
            raise ValueError(f"invalid result_type: '{result_type}'")

    def parse_map(self, obj):
        members = obj[1]
        return dict(zip(members.keys(), map(self.handle_type, members.values())))

    def parse_mapping(self, obj):
        return self.parse_map(obj)

    def parse_list(self, obj):
        return list(map(self.handle_type, obj[1]))

    def parse_listing(self, obj):
        return self.parse_list(obj)

    def parse_set(self, obj):
        return set(obj[1])

    def parse_duration(self, obj):
        _, value, unit = obj

        return Duration(value, unit)

    def parse_pair(self, obj):
        if self._pair_result_type == ResultType.DATACLASS:
            return Pair(obj[1], obj[2])
        elif self._pair_result_type == ResultType.NAMEDTUPLE:
            return self._namedtuple_cache["__Pair__"](obj[1], obj[2])
        elif self._pair_result_type == ResultType.LIST:
            return [obj[1], obj[2]]
        elif self._pair_result_type == ResultType.TUPLE:
            return (obj[1], obj[2])
        else:
            raise ValueError(f"ResultType '{self._pair_result_type}' not supported")

    def parse_datasize(self, obj):
        if self._datasize_result_type == ResultType.DATACLASS:
            return DataSize(obj[1], obj[2])
        elif self._datasize_result_type == ResultType.NAMEDTUPLE:
            return self._namedtuple_cache["__DataSize__"](obj[1], obj[2])
        elif self._datasize_result_type == ResultType.TUPLE:
            return (obj[1], obj[2])
        else:
            raise ValueError(f"ResultType '{self._datasize_result_type}' not supported")

    def parse_intseq(self, obj):
        if self._intseq_result_type == ResultType.DATACLASS:
            return IntSeq(obj[1], obj[2], obj[3])
        elif self._intseq_result_type == ResultType.RANGE:
            return range(obj[1], obj[2], obj[3])
        else:
            raise ValueError(f"ResultType '{self._intseq_result_type}' not supported")

    def parse_regex(self, obj):
        if self._datasize_result_type == ResultType.DATACLASS:
            return Regex(obj[1])
        elif self._datasize_result_type == ResultType.NAMEDTUPLE:
            return self._namedtuple_cache["__Regex__"](obj[1])
        else:
            raise ValueError(f"ResultType '{self._datasize_result_type}' not supported")

    def parse_class(self, obj):
        return

    def parse_typealias(self, obj):
        return

    def parse_property(self, obj):
        _, key, value = obj
        return {key: self.handle_type(value)}

    def parse_entry(self, obj):
        _, key, value = obj
        return {key: self.handle_type(value)}

    def parse_element(self, obj):
        _, index, value = obj
        return {index: self.handle_type(value)}



from pathlib import Path

import pytest

import pkll

base_path = Path("./tests/Fixtures")


def test_AnyType():
    file = base_path / "AnyType.pkl"
    _ = pkll.load(file)


def test_ApiTypes():
    file = base_path / "ApiTypes.pkl"
    _ = pkll.load(file)


def test_Classes():
    file = base_path / "Classes.pkl"
    _ = pkll.load(file)


def test_Collections():
    file = base_path / "Collections.pkl"
    _ = pkll.load(file)


def test_ExtendedModule():
    file = base_path / "ExtendedModule.pkl"
    _ = pkll.load(file)


def test_OpenModule():
    file = base_path / "OpenModule.pkl"
    with pytest.raises(Exception):
        _ = pkll.load(file)


def test_Poly():
    file = base_path / "Poly.pkl"
    _ = pkll.load(file)


def test_UnionTypes():
    file = base_path / "UnionTypes.pkl"
    _ = pkll.load(file)


def test_lib1():
    file = base_path / "lib1.pkl"
    _ = pkll.load(file)


def test_types():
    file = Path("./tests") / "types.pkl"
    _ = pkll.load(file)



# Code generated from Pkl module `ExtendingOpenClass`. DO NOT EDIT.
from enum import Enum
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll
from . import lib3_pkl

_NO_DEFAULT = object()

@dataclass
class MyClass(MyOpenClass):
    myBoolean: bool = _NO_DEFAULT

    myStr: str = "mystr"

    _registered_identifier: str = "ExtendingOpenClass#MyClass"

    def __post_init__(self):
        no_default_list = [("myBoolean", self.myBoolean)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

@dataclass
class MyOpenClass:
    myStr: str

    _registered_identifier: str = "ExtendingOpenClass#MyOpenClass"

@dataclass
class MyClass2(lib3_pkl.GoGoGo):
    myBoolean: bool = _NO_DEFAULT

    duck: Literal["quack"] = "quack"

    _registered_identifier: str = "ExtendingOpenClass#MyClass2"

    def __post_init__(self):
        no_default_list = [("myBoolean", self.myBoolean)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

@dataclass
class ExtendingOpenClass:
    res1: MyClass

    res2: MyClass2

    _registered_identifier: str = "ExtendingOpenClass"

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `ExtendingOpenClass.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())


from dataclasses import asdict, dataclass, field, is_dataclass
from typing import Dict, List, Optional, Union

from pkll.msgapi.code import (
    CODE_CLOSE_EVALUATOR,
    CODE_EVALUATE,
    CODE_EVALUATE_READ_MODULE_RESPONSE,
    CODE_EVALUATE_READ_RESPONSE,
    CODE_LIST_MODULES_RESPONSE,
    CODE_LIST_RESOURCES_RESPONSE,
    CODE_NEW_EVALUATOR,
)


@dataclass
class OutgoingMessage:
    def to_msg_obj(self):
        raise NotImplementedError("Subclasses must implement this method.")


def dataclass_to_dict(dc):
    if not is_dataclass(dc):
        return dc

    dd = {k: dataclass_to_dict(v) for k, v in asdict(dc).items()}
    return dd


def _filter_dict(dd):
    if not isinstance(dd, dict):
        return dd
    return {k: _filter_dict(v) for k, v in dd.items() if v is not None}


def format_message(msg: OutgoingMessage, code: int):
    # msg_dict = asdict(msg)
    # filtered_dict = {k: v for k, v in msg_dict.items() if v is not None}
    msg_dict = dataclass_to_dict(msg)
    filtered_dict = _filter_dict(msg_dict)
    msg_obj = [code, filtered_dict]
    return msg_obj


@dataclass
class ResourceReader:
    scheme: str
    has_hierarchical_uris: bool
    is_globbable: bool


@dataclass
class ModuleReader:
    scheme: str
    has_hierarchical_uris: bool
    is_globbable: bool
    is_local: bool


@dataclass
class Checksums:
    sha256: str


@dataclass
class PathElement:
    name: str
    is_directory: bool


@dataclass
class ProjectOrDependency:
    package_uri: Optional[str] = None
    type: str = ""
    project_file_uri: Optional[str] = None
    checksums: Optional[Checksums] = None
    dependencies: Dict[str, "ProjectOrDependency"] = field(default_factory=dict)


@dataclass
class ClientResourceReader:
    scheme: str
    hasHierarchicalUris: bool
    isGlobbable: bool


@dataclass
class ClientModuleReader:
    scheme: str
    hasHierarchicalUris: bool
    isGlobbable: bool
    isLocal: bool


@dataclass
class RemoteDependency:
    type: str = "remote"
    packageUri: Optional[str] = None
    checksums: Optional[Checksums] = None


@dataclass
class Project:
    projectFileUri: str
    packageUri: Optional[str] = None
    type: str = "local"
    dependencies: Dict[str, Union["Project", RemoteDependency]] = field(
        default_factory=dict
    )


@dataclass
class CreateEvaluator(OutgoingMessage):
    requestId: int
    allowedModules: Optional[List[str]] = field(default=None)
    allowedResources: Optional[List[str]] = field(default=None)
    clientModuleReaders: Optional[List[ClientModuleReader]] = field(default=None)
    clientResourceReaders: Optional[List[ClientResourceReader]] = field(default=None)
    modulePaths: Optional[List[str]] = field(default=None)
    env: Optional[Dict[str, str]] = field(default=None)
    properties: Optional[Dict[str, str]] = field(default=None)
    timeoutSeconds: Optional[int] = field(default=None)
    rootDir: Optional[str] = field(default=None)
    cacheDir: Optional[str] = field(default=None)
    outputFormat: Optional[str] = field(default=None)
    project: Optional[Project] = field(default=None)

    def to_msg_obj(self):
        return format_message(self, CODE_NEW_EVALUATOR)


@dataclass
class CloseEvaluator(OutgoingMessage):
    evaluatorId: int

    def to_msg_obj(self):
        return format_message(self, CODE_CLOSE_EVALUATOR)


@dataclass
class EvaluateRequest(OutgoingMessage):
    """
    A class to represent a request for evaluating a module or expression in Pkl.

    Attributes:
    - requestId: A number identifying this request.
    - evaluatorId: A number identifying the evaluator.
    - moduleUri: The absolute URI of the module to be evaluated.
    - moduleText: The module's contents. If None, Pkl will load the module at runtime.
    - expr: The Pkl expression to be evaluated within the module. If None, evaluates the whole module.
    """

    requestId: int
    evaluatorId: int
    moduleUri: str
    moduleText: Optional[str] = None
    expr: Optional[str] = None

    def to_msg_obj(self):
        return format_message(self, CODE_EVALUATE)


@dataclass
class EvaluatorReadResourceResponse(OutgoingMessage):
    requestId: int
    evaluatorId: int
    contents: Optional[bytes] = None
    error: Optional[str] = None

    def to_msg_obj(self):
        return format_message(self, CODE_EVALUATE_READ_RESPONSE)


@dataclass
class EvaluatorReadModuleResponse(OutgoingMessage):
    requestId: int
    evaluatorId: int
    contents: Optional[str] = None
    error: Optional[str] = None

    def to_msg_obj(self):
        return format_message(self, CODE_EVALUATE_READ_MODULE_RESPONSE)


@dataclass
class EvaluatorListResourcesResponse(OutgoingMessage):
    requestId: int
    evaluatorId: int
    pathElements: List[PathElement] = field(default_factory=list)
    error: Optional[str] = None

    def to_msg_obj(self):
        return format_message(self, CODE_LIST_RESOURCES_RESPONSE)


@dataclass
class EvaluatorListModulesResponse(OutgoingMessage):
    requestId: int
    evaluatorId: int
    pathElements: List[PathElement] = field(default_factory=list)
    error: Optional[str] = None

    def to_msg_obj(self):
        return format_message(self, CODE_LIST_MODULES_RESPONSE)



import argparse
import logging
from pathlib import Path
from urllib.request import pathname2url
import pkll

# Mock-up classes and functions to match the TypeScript functionality
# These would need to be implemented or adapted for your specific use case
# from src import newEvaluator, PreconfiguredOptions
# from generated import GeneratorSettings, load as loadGeneratorSettings
# from generate import generateTypescript

from generate import generate_python, GeneratorSettings

logger = logging.getLogger()
logging.basicConfig(level=logging.INFO, format="%(message)s")


def setup_logger(verbose):
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


def cli_handler(pkl_modules, settings_file_path, output_directory, dry_run, verbose):
    setup_logger(verbose)

    if not pkl_modules:
        logger.error("You must provide at least one file to evaluate.")
        return

    settings_file = settings_file_path or (Path.cwd() / "generator-settings.pkl")
    if not settings_file.exists():
        settings_file = None

    if settings_file:
        logger.info(f"Using settings file at {settings_file}")
    else:
        logger.info("No settings file found, using default settings.")

    if settings_file:
        settings = pkll.load(pathname2url(str(settings_file)))
    else:
        settings = GeneratorSettings(dry_run=dry_run, output_directory=output_directory)

    generate_python(pkl_modules, settings, verbose)


def main():
    parser = argparse.ArgumentParser(
        description="Generate TypeScript from PKL modules."
    )
    parser.add_argument(
        "pklModules",
        metavar="PklModule",
        type=str,
        nargs="+",
        help="Pkl module to evaluate",
    )
    parser.add_argument(
        "-s",
        "--settings-file",
        type=str,
        help="Path to the generator-settings.pkl file",
    )
    parser.add_argument(
        "-o",
        "--output-directory",
        type=str,
        help="Directory to write generated files into",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Evaluate the Pkl modules but do not write them anywhere",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable debug logging"
    )

    args = parser.parse_args()

    cli_handler(
        pkl_modules=args.pklModules,
        settings_file_path=Path(args.settings_file) if args.settings_file else None,
        output_directory=args.output_directory,
        dry_run=args.dry_run,
        verbose=args.verbose,
    )


if __name__ == "__main__":
    main()



import os
import platform
import stat
from pathlib import Path

import setuptools
from setuptools.command.install import install

BINARIES = {
    ("darwin", "64bit"): "pkl-macos-amd64",
    ("darwin", "aarch64"): "pkl-macos-aarch64",
    ("linux", "64bit"): "pkl-linux-amd64",
    ("linux", "aarch64"): "pkl-linux-aarch64",
    ("linux", "64bit", "alpine"): "pkl-alpine-linux-amd64",
}
PKL_VERSION = "0.25.2"
BASE_PATH = "https://github.com/apple/pkl/releases/download/"


def detect_system():
    os_name = platform.system().lower()
    arch, _ = platform.architecture()
    return os_name, arch


def is_alpine_linux():
    return os.path.isfile("/etc/alpine-release")


def download_binary(file, target_fp):
    import requests

    url = BASE_PATH + PKL_VERSION + "/" + file
    response = requests.get(url)
    with open(target_fp, "wb") as f:
        f.write(response.content)


def get_binary_path():
    os_name, arch = detect_system()
    if os_name == "linux" and is_alpine_linux():
        os_name = "alpine"
    binary_key = (os_name, arch)
    if binary_key == ("linux", "64bit") and is_alpine_linux():
        binary_key += ("alpine",)
    bin_file = BINARIES.get(binary_key)

    if bin_file is None:
        raise OSError("No compatible binary found for your system.")

    bin_parent_path = (Path("~/.pkl/bin/") / PKL_VERSION).expanduser()
    binary_path = bin_parent_path / bin_file

    if not binary_path.exists():
        binary_path.parent.mkdir(exist_ok=True, parents=True)
        download_binary(bin_file, binary_path)

    current_permissions = os.stat(binary_path).st_mode
    new_permissions = current_permissions | stat.S_IXUSR
    os.chmod(binary_path, new_permissions)

    return binary_path


class CustomInstallCommand(install):
    def run(self):
        get_binary_path()
        install.run(self)


setuptools.setup(
    cmdclass={"install": CustomInstallCommand},
    setup_requires=["requests"],
)



from dataclasses import dataclass, field
from typing import List, Optional

from pkll.msgapi.outgoing import PathElement


@dataclass
class ListResponse:
    pathElements: List[PathElement] = field(default_factory=list)
    error: Optional[str] = None


@dataclass
class ReadModuleResponse:
    contents: Optional[str] = None
    error: Optional[str] = None


@dataclass
class ReadResourceResponse:
    contents: Optional[bytes] = None
    error: Optional[str] = None


class Handler:
    def list_response(self, uri: str) -> ListResponse:
        raise NotImplementedError


class ModuleHandler(Handler):
    def read_response(self, uri: str) -> ReadModuleResponse:
        raise NotImplementedError


class ResourcesHandler(Handler):
    def read_response(self, uri: str) -> ReadResourceResponse:
        raise NotImplementedError



import argparse
import os
import sys
import tempfile
from pathlib import Path
from typing import Dict
from textwrap import dedent

import pkll


def file_exists(filepath):
    return Path(filepath).exists()


def do_find_project_dir(directory: Path) -> Path:
    if (directory / "PklProject").exists():
        return directory
    parent = directory.parent
    if parent == directory:
        return Path()
    return do_find_project_dir(parent)


def find_project_dir(project_dir_flag: str) -> Path:
    if project_dir_flag:
        return Path(project_dir_flag).resolve()
    cwd = Path.cwd()
    return do_find_project_dir(cwd)


def load_generator_settings(generator_settings_path: str, project_dir_flag: str):
    project_dir = find_project_dir(project_dir_flag)
    if project_dir:
        print(f"Project directory found: {project_dir}")
    else:
        print("No specific project directory found, using current working directory.")

    if generator_settings_path:
        settings_path = Path(generator_settings_path)
    else:
        settings_path = project_dir / "generator-settings.pkl"
        if not settings_path.exists():
            print("Generator settings file not found. Using default settings...")
            raise NotImplementedError

    print(f"Loading settings from: {settings_path}")

    config = pkll.load(settings_path.absolute().as_uri(), allowedResources=["file:"])
    return config


def generate_dry_run(tmp_file_path: str, output_path: str, settings) -> None:
    #filenames = evaluate_pkl("output.files.toMap().keys.toList()", tmp_file_path)
    filenames = pkll.load(Path(tmp_file_path).as_uri(), expr="output.files.toMap().keys.toList()")
    
    print("Dry run; printing filenames but not writing files to disk")
    for filename in filenames:
        if settings.basePath and filename.startswith(settings.basePath):
            #filename = filename[len(settings.basePath) :]
            filename = Path(filename).relative_to(settings.basePath)
        else:
            continue
        print(Path(output_path) / filename)


def generate_python(
    pkl_module_path: str, settings, output_path: str, dry_run: bool
) -> None:
    module_to_evaluate = dedent(
        f"""\
        amends "{Path(settings.generatorScriptPath).absolute()}"
        import "{Path(pkl_module_path).absolute()}" as theModule

        moduleToGenerate = theModule
    """
    )
    module_to_evaluate = dedent(
        """\
        import "{}" as Generator
        import "{}" as theModule

        local gen = new Generator {{
          codegenSettings {{
            packageMappings {{
                {}
            }}
            {}
            structTags {{
                {}
            }}
          }}
          moduleToGenerate = theModule
        }}
        output = gen.output\
    """.format(
            Path(settings.generatorScriptPath).absolute(),
            Path(pkl_module_path).absolute(),
            "\\n".join(f'["{k}"] = "{v}"' for k, v in settings.packageMappings),
            f'basePath = "{settings.basePath}"' if settings.basePath else "",
            "\\n".join(f'["{k}"] = "{v}"' for k, v in settings.structTags),
        )
    )
    module_to_evaluate = dedent(
        """\
        amends "{}"
        import "{}" as theModule

        codegenSettings = new {{
            packageMappings {{
                {}
            }}
            {}
            structTags {{
                {}
            }}
        }}
        moduleToGenerate = theModule
    """.format(
            Path(settings.generatorScriptPath).absolute(),
            Path(pkl_module_path).absolute(),
            "\\n".join(f'["{k}"] = "{v}"' for k, v in settings.packageMappings),
            f'basePath = "{settings.basePath}"' if settings.basePath else "",
            "\\n".join(f'["{k}"] = "{v}"' for k, v in settings.structTags),
        )
    )

    with tempfile.NamedTemporaryFile(suffix=".pkl") as tmp_file:
        tmp_file.write(module_to_evaluate.encode())
        tmp_file.flush()
        print("module tmp file:", tmp_file.name)

        if dry_run or settings.dryRun:
            generate_dry_run(tmp_file.name, output_path, settings)
            return

        config = pkll.load(
            Path(tmp_file.name).as_uri(),
            expr="output",
            allowedResources=["env:", "prop:", "package:", "projectpackage:", "file:"],
            debug=True,
        )
        print(config)
    breakpoint()


def main():
    # Create the parser
    parser = argparse.ArgumentParser(
        description="Generates Python bindings for a Pkl module."
    )

    # Positional arguments
    parser.add_argument("module", help="The Pkl module to process.")

    # Optional arguments
    parser.add_argument(
        "--generator-settings",
        default="",
        help="The path to a generator settings file.",
    )
    parser.add_argument(
        "--output-path",
        default="",
        help="The output directory to write generated sources into.",
    )
    parser.add_argument(
        "--base-path",
        default="",
        help="The base path used to determine relative output paths.",
    )
    parser.add_argument(
        "--generate-script",
        default="",
        help="The Generate.pkl script to use for code generation.",
    )
    parser.add_argument(
        "--project-dir",
        default="",
        help="The project directory to load dependency and evaluator settings from.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Print out the names of the files that will be generated, but don't write any files.",
    )

    # Flags
    parser.add_argument(
        "--suppress-format-warning",
        action="store_true",
        help="Suppress warnings around formatting issues.",
    )
    parser.add_argument(
        "--version", action="store_true", help="Print the version of the tool and exit."
    )

    # Repeatable arguments
    parser.add_argument(
        "--mapping",
        action="append",
        nargs=2,
        metavar=("PKL_MODULE_NAME", "PYTHON_PACKAGE_NAME"),
        help="Mapping of a Pkl module name to a Python package name. This option can be repeated.",
    )
    parser.add_argument(
        "--allowed-modules",
        action="append",
        help="URI patterns that determine which modules can be loaded and evaluated. Can be repeated.",
    )
    parser.add_argument(
        "--allowed-resources",
        action="append",
        help="URI patterns that determine which resources can be read and evaluated. Can be repeated.",
    )

    # Parse the arguments
    args = parser.parse_args()

    if args.version:
        print("Version placeholder")  # Replace with actual version logic
        sys.exit()

    settings = load_generator_settings(args.generator_settings, args.project_dir)

    if not args.output_path:
        output_path = os.getcwd()
    else:
        output_path = args.output_path

    generate_python(args.module, settings, output_path, args.dry_run)


if __name__ == "__main__":
    main()



# Code generated from Pkl module `com.example.Simple`. DO NOT EDIT.
from enum import Enum
from typing import Dict, List, Literal, Optional, Union
from dataclasses import dataclass
import pkll

_NO_DEFAULT = object()

@dataclass
class Person:
    # The name of the person
    the_name: str

    # Some name that matches a keyword
    enum: str

    _registered_identifier: str = "com.example.Simple#Person"

@dataclass
class ThePerson(Person):
    the: str = _NO_DEFAULT

    _registered_identifier: str = "com.example.Simple#ThePerson"

    def __post_init__(self):
        no_default_list = [("the", self.the)]
        for n, x in no_default_list:
            if x is _NO_DEFAULT:
                raise TypeError(f"__init__ missing 1 required argument: '{n}'")

@dataclass
class OpenClassExtendingOpenClass:
    someOtherProp: Optional[bool]

    _registered_identifier: str = "com.example.Simple#OpenClassExtendingOpenClass"

@dataclass
class ClassWithReallyLongConstructor:
    theProperty1: str

    theProperty2: str

    theProperty3: str

    theProperty4: str

    theProperty5: str

    theProperty6: str

    _registered_identifier: str = "com.example.Simple#ClassWithReallyLongConstructor"

@dataclass
class com_example_Simple:
    # This is truly a person.
    person: Person

    _registered_identifier: str = "com.example.Simple"

    @classmethod
    def load_pkl(cls, source: str):
        # Load the Pkl module at the given source and evaluate it into `com_example_Simple.Module`.
        # - Parameter source: The source of the Pkl module.
        config = pkll.load(source)
        return cls(**config._asdict())